/*
 * livepatch_bsc1199602
 *
 * Fix for CVE-2022-30594, bsc#1199602
 *
 *  Upstream commit:
 *  ee1fee900537 ("ptrace: Check PTRACE_O_SUSPEND_SECCOMP permission on
 *                 PTRACE_SEIZE")
 *
 *  SLE12-SP3 commit:
 *  2c177a4fe9df8fd8ffe0bc075c842e4bb53ce3c6
 *
 *  SLE12-SP4, SLE15 and SLE15-SP1 commit:
 *  26d8e0b8e00657d34be0e8b628a72903b17723b2
 *
 *  SLE12-SP5 commit:
 *  9eb132f4c5dd0e8c7ac813b70e65f4eb26988d67
 *
 *  SLE15-SP2 commit:
 *  fd4d93d4cb8ab024512df2f32948d63f4dcb4549
 *
 *  SLE15-SP3 commit
 *  daaf8a2d25679dc82d099a32b6cd1e14b4f1e546
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from kernel/ptrace.c */
#include <linux/capability.h>
#include <linux/export.h>
#include <linux/sched.h>

/* klp-ccp: from include/linux/sched.h */
static struct task_struct *(*klpe_find_get_task_by_vpid)(pid_t nr);

static int (*klpe_wake_up_state)(struct task_struct *tsk, unsigned int state);

static unsigned long (*klpe_wait_task_inactive)(struct task_struct *, long match_state);

/* klp-ccp: from kernel/ptrace.c */
#include <linux/sched/mm.h>
#include <linux/sched/task.h>

/* klp-ccp: from include/linux/sched/task.h */
static rwlock_t (*klpe_tasklist_lock);

/* klp-ccp: from kernel/ptrace.c */
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/ptrace.h>

/* klp-ccp: from include/linux/sched/jobctl.h */
static bool (*klpe_task_set_jobctl_pending)(struct task_struct *task, unsigned long mask);

/* klp-ccp: from include/linux/sched/signal.h */
static void (*klpe_signal_wake_up_state)(struct task_struct *t, unsigned int state);

/* klp-ccp: from include/linux/ptrace.h */
static long (*klpe_arch_ptrace)(struct task_struct *child, long request,
			unsigned long addr, unsigned long data);

static void (*klpe___ptrace_link)(struct task_struct *child,
			  struct task_struct *new_parent,
			  const struct cred *ptracer_cred);

/* klp-ccp: from include/linux/security.h */
#ifdef CONFIG_SECURITY

static int (*klpe_security_ptrace_traceme)(struct task_struct *parent);

#else /* CONFIG_SECURITY */
#error "klp-ccp: non-taken branch"
#endif	/* CONFIG_SECURITY */

/* klp-ccp: from kernel/ptrace.c */
#include <linux/signal.h>
#include <linux/audit.h>

/* klp-ccp: from include/linux/audit.h */
#ifdef CONFIG_AUDITSYSCALL

static void (*klpe___audit_ptrace)(struct task_struct *t);

static inline void klpr_audit_ptrace(struct task_struct *t)
{
	if (unlikely(!audit_dummy_context()))
		(*klpe___audit_ptrace)(t);
}

#else /* CONFIG_AUDITSYSCALL */
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_AUDITSYSCALL */

/* klp-ccp: from kernel/ptrace.c */
#include <linux/pid_namespace.h>

/* klp-ccp: from kernel/ptrace.c */
#include <linux/uaccess.h>

/* klp-ccp: from include/linux/cn_proc.h */
#ifdef CONFIG_PROC_EVENTS

static void (*klpe_proc_ptrace_connector)(struct task_struct *task, int which_id);

#else
#error "klp-ccp: non-taken branch"
#endif	/* CONFIG_PROC_EVENTS */

/* klp-ccp: from kernel/ptrace.c */
#include <linux/compat.h>
#include <linux/sched/signal.h>
#include <asm/syscall.h>	/* for syscall_get_* */

static void klpr_ptrace_link(struct task_struct *child, struct task_struct *new_parent)
{
	(*klpe___ptrace_link)(child, new_parent, current_cred());
}

static bool ptrace_freeze_traced(struct task_struct *task)
{
	bool ret = false;

	/* Lockless, nobody but us can set this flag */
	if (task->jobctl & JOBCTL_LISTENING)
		return ret;

	spin_lock_irq(&task->sighand->siglock);
	if (task_is_traced(task) && !__fatal_signal_pending(task)) {
		task->state = __TASK_TRACED;
		ret = true;
	}
	spin_unlock_irq(&task->sighand->siglock);

	return ret;
}

static void klpr_ptrace_unfreeze_traced(struct task_struct *task)
{
	if (task->state != __TASK_TRACED)
		return;

	WARN_ON(!task->ptrace || task->parent != current);

	/*
	 * PTRACE_LISTEN can allow ptrace_trap_notify to wake us up remotely.
	 * Recheck state under the lock to close this race.
	 */
	spin_lock_irq(&task->sighand->siglock);
	if (task->state == __TASK_TRACED) {
		if (__fatal_signal_pending(task))
			(*klpe_wake_up_state)(task, __TASK_TRACED);
		else
			task->state = TASK_TRACED;
	}
	spin_unlock_irq(&task->sighand->siglock);
}

static int klpr_ptrace_check_attach(struct task_struct *child, bool ignore_state)
{
	int ret = -ESRCH;

	/*
	 * We take the read lock around doing both checks to close a
	 * possible race where someone else was tracing our child and
	 * detached between these two checks.  After this locked check,
	 * we are sure that this is our traced child and that can only
	 * be changed by us so it's not changing right after this.
	 */
	read_lock(&(*klpe_tasklist_lock));
	if (child->ptrace && child->parent == current) {
		WARN_ON(child->state == __TASK_TRACED);
		/*
		 * child->sighand can't be NULL, release_task()
		 * does ptrace_unlink() before __exit_signal().
		 */
		if (ignore_state || ptrace_freeze_traced(child))
			ret = 0;
	}
	read_unlock(&(*klpe_tasklist_lock));

	if (!ret && !ignore_state) {
		if (!(*klpe_wait_task_inactive)(child, __TASK_TRACED)) {
			/*
			 * This can only happen if may_ptrace_stop() fails and
			 * ptrace_stop() changes ->state back to TASK_RUNNING,
			 * so we should not worry about leaking __TASK_TRACED.
			 */
			WARN_ON(child->state == __TASK_TRACED);
			ret = -ESRCH;
		}
	}

	return ret;
}

static int (*klpe___ptrace_may_access)(struct task_struct *task, unsigned int mode);

/* New. */
static int klpp_check_ptrace_options(unsigned long data)
{
	if (data & ~(unsigned long)PTRACE_O_MASK)
		return -EINVAL;

	if (unlikely(data & PTRACE_O_SUSPEND_SECCOMP)) {
		if (!IS_ENABLED(CONFIG_CHECKPOINT_RESTORE) ||
		    !IS_ENABLED(CONFIG_SECCOMP))
			return -EINVAL;

		if (!capable(CAP_SYS_ADMIN))
			return -EPERM;

		if (seccomp_mode(&current->seccomp) != SECCOMP_MODE_DISABLED ||
		    current->ptrace & PT_SUSPEND_SECCOMP)
			return -EPERM;
	}
	return 0;
}

static int klpp_ptrace_attach(struct task_struct *task, long request,
			 unsigned long addr,
			 unsigned long flags)
{
	bool seize = (request == PTRACE_SEIZE);
	int retval;

	retval = -EIO;
	if (seize) {
		if (addr != 0)
			goto out;
		if (flags & ~(unsigned long)PTRACE_O_MASK)
			goto out;
		/*
		 * Fix CVE-2022-30594
		 *  +3 lines
		 */
		retval = klpp_check_ptrace_options(flags);
		if (retval)
			return retval;
		flags = PT_PTRACED | PT_SEIZED | (flags << PT_OPT_FLAG_SHIFT);
	} else {
		flags = PT_PTRACED;
	}

	klpr_audit_ptrace(task);

	retval = -EPERM;
	if (unlikely(task->flags & PF_KTHREAD))
		goto out;
	if (same_thread_group(task, current))
		goto out;

	/*
	 * Protect exec's credential calculations against our interference;
	 * SUID, SGID and LSM creds get determined differently
	 * under ptrace.
	 */
	retval = -ERESTARTNOINTR;
	if (mutex_lock_interruptible(&task->signal->cred_guard_mutex))
		goto out;

	task_lock(task);
	retval = (*klpe___ptrace_may_access)(task, PTRACE_MODE_ATTACH_REALCREDS);
	task_unlock(task);
	if (retval)
		goto unlock_creds;

	write_lock_irq(&(*klpe_tasklist_lock));
	retval = -EPERM;
	if (unlikely(task->exit_state))
		goto unlock_tasklist;
	if (task->ptrace)
		goto unlock_tasklist;

	if (seize)
		flags |= PT_SEIZED;
	task->ptrace = flags;

	klpr_ptrace_link(task, current);

	/* SEIZE doesn't trap tracee on attach */
	if (!seize)
		send_sig_info(SIGSTOP, SEND_SIG_PRIV, task);

	spin_lock(&task->sighand->siglock);

	/*
	 * If the task is already STOPPED, set JOBCTL_TRAP_STOP and
	 * TRAPPING, and kick it so that it transits to TRACED.  TRAPPING
	 * will be cleared if the child completes the transition or any
	 * event which clears the group stop states happens.  We'll wait
	 * for the transition to complete before returning from this
	 * function.
	 *
	 * This hides STOPPED -> RUNNING -> TRACED transition from the
	 * attaching thread but a different thread in the same group can
	 * still observe the transient RUNNING state.  IOW, if another
	 * thread's WNOHANG wait(2) on the stopped tracee races against
	 * ATTACH, the wait(2) may fail due to the transient RUNNING.
	 *
	 * The following task_is_stopped() test is safe as both transitions
	 * in and out of STOPPED are protected by siglock.
	 */
	if (task_is_stopped(task) &&
	    (*klpe_task_set_jobctl_pending)(task, JOBCTL_TRAP_STOP | JOBCTL_TRAPPING))
		(*klpe_signal_wake_up_state)(task, __TASK_STOPPED);

	spin_unlock(&task->sighand->siglock);

	retval = 0;
unlock_tasklist:
	write_unlock_irq(&(*klpe_tasklist_lock));
unlock_creds:
	mutex_unlock(&task->signal->cred_guard_mutex);
out:
	if (!retval) {
		/*
		 * We do not bother to change retval or clear JOBCTL_TRAPPING
		 * if wait_on_bit() was interrupted by SIGKILL. The tracer will
		 * not return to user-mode, it will exit and clear this bit in
		 * __ptrace_unlink() if it wasn't already cleared by the tracee;
		 * and until then nobody can ptrace this task.
		 */
		wait_on_bit(&task->jobctl, JOBCTL_TRAPPING_BIT, TASK_KILLABLE);
		(*klpe_proc_ptrace_connector)(task, PTRACE_ATTACH);
	}

	return retval;
}

static int klpr_ptrace_traceme(void)
{
	int ret = -EPERM;

	write_lock_irq(&(*klpe_tasklist_lock));
	/* Are we already being traced? */
	if (!current->ptrace) {
		ret = (*klpe_security_ptrace_traceme)(current->parent);
		/*
		 * Check PF_EXITING to ensure ->real_parent has not passed
		 * exit_ptrace(). Otherwise we don't report the error but
		 * pretend ->real_parent untraces us right after return.
		 */
		if (!ret && !(current->real_parent->flags & PF_EXITING)) {
			current->ptrace = PT_PTRACED;
			klpr_ptrace_link(current, current->real_parent);
		}
	}
	write_unlock_irq(&(*klpe_tasklist_lock));

	return ret;
}

#ifndef arch_ptrace_attach
#define arch_ptrace_attach(child)	do { } while (0)
#else
#error "klp-ccp: non-taken branch"
#endif

#include <linux/syscalls.h>

__SYSCALL_DEFINEx(4, _klpp_ptrace, long, request, long, pid, unsigned long, addr,
		unsigned long, data)
{
	struct task_struct *child;
	long ret;

	if (request == PTRACE_TRACEME) {
		ret = klpr_ptrace_traceme();
		if (!ret)
			arch_ptrace_attach(current);
		goto out;
	}

	child = (*klpe_find_get_task_by_vpid)(pid);
	if (!child) {
		ret = -ESRCH;
		goto out;
	}

	if (request == PTRACE_ATTACH || request == PTRACE_SEIZE) {
		ret = klpp_ptrace_attach(child, request, addr, data);
		/*
		 * Some architectures need to do book-keeping after
		 * a ptrace attach.
		 */
		if (!ret)
			arch_ptrace_attach(child);
		goto out_put_task_struct;
	}

	ret = klpr_ptrace_check_attach(child, request == PTRACE_KILL ||
				  request == PTRACE_INTERRUPT);
	if (ret < 0)
		goto out_put_task_struct;

	ret = (*klpe_arch_ptrace)(child, request, addr, data);
	if (ret || request != PTRACE_DETACH)
		klpr_ptrace_unfreeze_traced(child);

 out_put_task_struct:
	put_task_struct(child);
 out:
	return ret;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1199602.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "__audit_ptrace", (void *)&klpe___audit_ptrace },
	{ "__ptrace_link", (void *)&klpe___ptrace_link },
	{ "__ptrace_may_access", (void *)&klpe___ptrace_may_access },
	{ "arch_ptrace", (void *)&klpe_arch_ptrace },
	{ "find_get_task_by_vpid", (void *)&klpe_find_get_task_by_vpid },
	{ "proc_ptrace_connector", (void *)&klpe_proc_ptrace_connector },
	{ "security_ptrace_traceme", (void *)&klpe_security_ptrace_traceme },
	{ "signal_wake_up_state", (void *)&klpe_signal_wake_up_state },
	{ "task_set_jobctl_pending", (void *)&klpe_task_set_jobctl_pending },
	{ "tasklist_lock", (void *)&klpe_tasklist_lock },
	{ "wait_task_inactive", (void *)&klpe_wait_task_inactive },
	{ "wake_up_state", (void *)&klpe_wake_up_state },
};

int livepatch_bsc1199602_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
